/* babit.h */

#ifndef BABIT_H_INCLUDED
#define BABIT_H_INCLUDED

/* If you have only one 'Babited' object, you may
   predefine symbol 'BABITED_OBJECT'
   like this:

   struct YourOwnStruct;
   #define BABITED_OBJECT struct YourOwnStruct
   #include "babit.h"

   typedef struct YourOwnStruct {
       ...
       BabitConnector bc;
       ...
   } YourOwnStruct;
*/

#ifndef BABITED_OBJECT
#define BABITED_OBJECT void
#endif

/* You can predefine symbol BABIT_KEY before including babit.h */

#ifndef BABIT_KEY
#define BABIT_KEY void
#endif

#ifndef BABIT_OFFSET /* You can pre-define this to 'long' */
#define BABIT_OFFSET int
#endif

typedef BABIT_OFFSET BABIT_ROOT;

typedef struct BabitConnector {
/*	 BABITED_OBJECT *Left;
	 BABITED_OBJECT *Right;
	 BABITED_OBJECT *Father; */
	 BABIT_OFFSET Left;
	 BABIT_OFFSET Right;
	 BABIT_OFFSET Father;
	 int  Balance;
} BabitConnector;

typedef int BabitCmpFun (const BABITED_OBJECT *obj1,
                         const BABITED_OBJECT *obj2);
typedef int BabitFndFun (const BABIT_KEY *key, 
                         const BABITED_OBJECT *obj);

typedef struct BabitControlBlock {
         int options;            /* set this to zero */
	 unsigned offset;        /* offset of connector in structure */
	 BabitCmpFun *cmpfun;    /* compares two structures' key */
	 BabitFndFun *fndfun;    /* comperes a key with a structure's key */
} BabitControlBlock;

#ifndef offsetof
#define offsetof(T,F) ((unsigned)&(((T *)0)->F))
#endif

#define BabitInit(rootp) (*(rootp) = 0)

#define BabitPtrToOffs(rootp,ptr) ((ptr) ? ((char *)(ptr)-(char *)(rootp)) : 0)
#define BabitOffsToPtr(rootp,off) ((BABITED_OBJECT *)\
                                  ((off) ? (char *)(rootp)+(off) : 0))

extern int BabitCheck (const BABIT_ROOT *root, int *dom, int *he,
		       const BabitControlBlock *bcb);

extern BABITED_OBJECT *BabitInsert (BABIT_ROOT *root,
				    BABITED_OBJECT *object,
				    const BabitControlBlock *bcb);

extern int   BabitDelete (BABIT_ROOT *root,
			  BABITED_OBJECT *object,
			  const BabitControlBlock *bcb);

extern BABITED_OBJECT *BabitFind (const BABIT_ROOT *root, const BABIT_KEY *key,
				  const BabitControlBlock *bcb);

/* FindExtended:
   returns an object for which fndfun (key, object)
   
    <  value   if option == BABIT_FIND_LT  (L)
    <= value   if option == BABIT_FIND_LE  (L)
    == value   if option == BABIT_FIND_EQ  (C)
    >= value   if option == BABIT_FIND_GE  (R)
    >  value   if option == BABIT_FIND_GT  (R)
    
    L: finds the leftmost (least) object for which the condition is true
    R: finds the rightmot (greatest) object for which the contidion is true
    C: finds the closest object for which the condition is true
*/    

#define BABIT_FIND_LT (-2)
#define BABIT_FIND_LE (-1)
#define BABIT_FIND_EQ   0
#define BABIT_FIND_GE   1
#define BABIT_FIND_GT   2

extern BABITED_OBJECT *BabitFindEx (const BABIT_ROOT *root,
				    const BABIT_KEY *key,
				    int option, int value,
				    const BabitControlBlock *bcb);

extern int   BabitReplace (BABIT_ROOT *root,
			   BABITED_OBJECT *old, BABITED_OBJECT *new,
			   const BabitControlBlock *bcb);

#define BABIT_GET_MIN (-1)
#define BABIT_GET_MED   0
#define BABIT_GET_MAX   1

extern BABITED_OBJECT *BabitGetMin (const BABIT_ROOT *root,
                                    int dir,
				    const BabitControlBlock *bcb);  
			
#define BABIT_GET_NEXT 1
#define BABIT_GET_PREV (-1)

extern BABITED_OBJECT *BabitGetNext (const BABITED_OBJECT *object, int dir,
				     const BabitControlBlock *bcb);
		
#define BABIT_WALK_PREORDER  1
#define BABIT_WALK_INORDER   2
#define BABIT_WALK_POSTORDER 4
#define BABIT_WALK_ALLORDER  7

#define BABIT_WALK_EOF -1
#define BABIT_WALK_ROOT 0 /* starting at root */
#define BABIT_WALK_DL   1 /* coming here from my father, I am the left child */
#define BABIT_WALK_DR   2 /* coming here from my father, I am the right child */
#define BABIT_WALK_UL   3 /* coming here from my left child */
#define BABIT_WALK_UR   4 /* coming here from my right child */

typedef struct BabitWalkStruct {
    int first;                    /* set to 1 on first call */
    int options;                  /* BABIT_WALK_xORDER */
/*  BABITED_OBJECT *obj;        *//* output */
    BABIT_OFFSET  obj;            /* output */
    int depth;			  /* output */
    int dir;                      /* output: BABIR_WALK_DL ... */
} BabitWalkStruct;

extern BABITED_OBJECT *BabitWalk (const BABIT_ROOT *root,
                                  BabitWalkStruct *bw,
				  const BabitControlBlock *bcb);
				  
#endif
